import os
import pickle

import click
import numpy as np
from torch import tensor
from sbi.inference import SNPE, SNLE

from pymob.utils.store_file import import_package, prepare_casestudy
from pymob.utils import help


def round_to_multiple(number, multiple):
    return multiple * round(number / multiple)


@click.command()
@click.option("-c", "--case_study", type=str, default="core_daphnia", 
              help=help.case_study)
@click.option("-s", "--scenario", type=str, default="expo_control", 
              help=help.scenario)
@click.option("-i", "--inferer_engine", type=str, default="SNPE")
@click.option("-l", "--split", type=int, default=None)
@click.option("-b", "--training_batch_size", type=int, default=None)
def main(case_study, scenario, inferer_engine, split, training_batch_size):
    # parse sbi config file
    config = prepare_casestudy((case_study, scenario), "sbi.cfg")

    output = config["case-study"].get("output")
    mod = import_package(package_path=config["case-study"]["package"])
    prior = getattr(mod.prior, config["sbi"]["prior"])
 

    # concatenate files or read existing simulation file
    theta_path = os.path.join(output, "theta.txt")
    theta = tensor(np.loadtxt(theta_path, dtype=np.float32))

    x_path = os.path.join(output, "x.txt")
    x = tensor(np.loadtxt(x_path, dtype=np.float32))
    # Train estimator ==============================================================
    # Create inference object: choose method and estimator

    # nsf estimator implements normalizing spline functions (nsf), which 
    # grant the estimator more freedom to fit any distribution
    if split is None:
        split = theta.shape[0]

    if training_batch_size is None:
        training_batch_size = max(1, round_to_multiple(split / 100, 10))

    # initialize density estimator
    if inferer_engine == "SNPE":
        inferer = SNPE(prior, density_estimator="nsf", device="cpu")
        fname = "SNPE_neural_posterior_estimator"  
        inferer = inferer.append_simulations(theta[0:split], x[0:split]) 
        density_estimator = inferer.train(
            training_batch_size=training_batch_size
        )  # Lots of training settings.

        # Build posterior using trained density estimator
        posterior = inferer.build_posterior(density_estimator)

    elif inferer_engine == "SNLE":
        inferer = SNLE(prior, density_estimator="nsf", device="cpu")  
        fname = "SNLE_neural_likelihood_estimator"  
        inferer = inferer.append_simulations(theta[0:split], x[0:split]) 
        density_estimator = inferer.train(
            training_batch_size=training_batch_size
        )  # Lots of training settings.

        # # Build posterior using trained density estimator
        # potential_fn, theta_transform_ = likelihood_estimator_based_potential(
        #     likelihood_estimator=density_estimator, prior=prior, x_o=None
        # )

        # theta_transform = IndependentTransform(prior.transforms[0], 1).inv

        # posterior = MCMCPosterior(
        #     potential_fn=potential_fn,
        #     theta_transform=theta_transform_,
        #     proposal=prior,
        #     method="nuts",
        #     device="cpu",
        #     x_shape=inferer._x_shape
        # )

        posterior = inferer.build_posterior(density_estimator)


    else:
        raise NotImplementedError(f"{inferer_engine} is not implemented choose SNPE or SNLE")

    # with 1 mio datapoints, the network takes approx. 1-2 minutes for one epoch
    # if the inferer takes approx 100-200 epochs to converge it will take max 400 
    # minutes ~ 6-7 hours.
    # CAUTION: Has to be adapted when training set size changes.

    # create dir if not exists
    dirname = os.path.join(output, inferer_engine.lower())
    os.makedirs(dirname, exist_ok=True)

    with open(os.path.join(dirname, f"{fname}.pkl"), "wb") as handle:
        pickle.dump(posterior, handle)

    with open(os.path.join(dirname, f"{inferer_engine}_inferer_logs.pkl"), "wb") as handle:
        pickle.dump(inferer.summary, handle)

if __name__ == "__main__":
    main()



import pytest
from pathlib import Path
from click.testing import CliRunner
import tempfile
from pymob.simulation import SimulationBase, Config
from pymob.sim.config import DataVariable, Datastructure, configure
from pymob.sim.parameters import Param, RandomVariable, Expression
from pymob.solvers.scipy import solve_ivp_1d
from sympy import Function
import xarray as xr
import numpy as np
import os

scenario = "case_studies/lotka_volterra_case_study/scenarios/test_scenario_scripting_api"

def test_simulation():
    sim = SimulationBase()
    sim.config.case_study.name = "lotka_volterra_case_study"
    sim.config.case_study.scenario = "test_scenario_scripting_api"
    sim.config.case_study.observations = "simulated_data.nc"
    sim.config.case_study.data_path = None

    # load data before specifying dims
    sim.config.case_study.data = os.path.abspath("case_studies/lotka_volterra_case_study/data")
    sim.observations = xr.load_dataset(sim.config.input_file_paths[0])    

    # try wrong specification
    threw_error = None
    try:
        sim.config.data_structure.rabbits = DataVariable(dimensions=["hash"])
        sim.config.data_structure.wolves = DataVariable(dimensions=["time"])
        sim.observations = xr.load_dataset(sim.config.input_file_paths[0])
        threw_error = False
    except KeyError:
        threw_error = True

    assert threw_error

    sim.validate()

    sim.config.data_structure.rabbits = DataVariable(dimensions=["time"])
    sim.config.data_structure.wolves = DataVariable(dimensions=["time"])
    
    # load data by providing an absolute path
    sim.config.case_study.data = os.path.abspath("case_studies/lotka_volterra_case_study/data")
    sim.observations = xr.load_dataset(sim.config.input_file_paths[0])    

    # load data by providing a relative path
    sim.config.case_study.data = "case_studies/lotka_volterra_case_study/data"
    sim.observations = xr.load_dataset(sim.config.input_file_paths[0])    
    
    # load data by providing no path (the default 'data' directory in the case study)
    sim.config.case_study.data = None
    sim.observations = xr.load_dataset(sim.config.input_file_paths[0])    

    sim.config.case_study.output = None

    # resetting the path is important in case other case studies were imported
    # in the same testing session
    # sim.config.import_casestudy_modules(reset_path=True)
    
    def lotka_volterra(t, y, alpha, beta, gamma, delta):
        prey, predator = y
        dprey_dt = alpha * prey - beta * prey * predator
        dpredator_dt = delta * prey * predator - gamma * predator
        return dprey_dt, dpredator_dt
    
    sim.model = lotka_volterra
    sim.solver = solve_ivp_1d
    sim.setup()
    sim.config.save(
        fp=f"{scenario}/test_settings.cfg",
        force=True, 
    )
    sim.save_observations(filename=sim.config.case_study.observations, force=True)

def test_load_generated_settings():
    sim = SimulationBase(f"{scenario}/test_settings.cfg")
    assert sim.config.case_study.name == "lotka_volterra_case_study"
    assert sim.config.case_study.scenario == "test_scenario_scripting_api"
    assert sim.config.case_study.package == "case_studies"
    assert sim.config.case_study.data == None
    assert Path(sim.config.case_study.data_path) == Path("case_studies/lotka_volterra_case_study/data")
    assert sim.config.case_study.output == None
    assert Path(sim.config.case_study.output_path) == \
        Path("case_studies/lotka_volterra_case_study/results/test_scenario_scripting_api")

def test_load_interpolated_settings():
    sim = SimulationBase(f"{scenario}/interp_settings.cfg")
    expected_output = \
        "./case_studies/lotka_volterra_case_study/results/test_scenario_scripting_api"
    assert sim.config.case_study.output == expected_output



def test_standalone_casestudy():
    wd = os.getcwd()
    case_study_name = "lotka_volterra_case_study_standalone"
    root = os.path.join(str(tempfile.tempdir), case_study_name)
    os.mkdir(root)
    os.chdir(root)
    
    # this is the syntax for setting up a standalone case study
    # currently root cannot be set with the config backend, but needs
    # to be specified with `chdir`
    sim = SimulationBase()
    sim.config.case_study.name = "."
    sim.config.case_study.scenario = "test_scenario_standalone"
    sim.config.case_study.package = "."

    os.makedirs(sim.config.case_study.output_path)
    os.makedirs(sim.config.case_study.data_path)
    os.makedirs(sim.config.case_study.scenario_path)
    sim.config.save(force=True)

    # test if all files exist and remove test directory
    os.chdir(wd)
    file_structure = [
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone",
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone/data",
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone/results",
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone/results/test_scenario_standalone",
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone/scenarios",
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone/scenarios/test_scenario_standalone",
        f"{tempfile.tempdir}/lotka_volterra_case_study_standalone/scenarios/test_scenario_standalone/settings.cfg",
    ]
    
    for p in reversed(file_structure):
        assert os.path.exists(p)
        if os.path.isdir(p):
            os.rmdir(p)
        else:
            os.remove(p)

def test_parameter_parsing():
    config = Config()

    io = "value=1.0 dims=[] min=0.0 max=3.0 hyper=False free=True"

    # test scripting input
    test = Param(value=1.0, min=0.0, max=3.0)
    config.model_parameters.test = test

    # test dict input
    config.model_parameters.test = test.model_dump(exclude_none=True)
    assert config.model_parameters.test == test # type: ignore

    # test validation
    config.model_parameters.test = io
    assert config.model_parameters.test == test # type: ignore

    # test serialization
    serialized = config.model_parameters.model_dump(mode="json")
    assert serialized == {"test": io}


def test_parameter_array():
    config = Config()

    io = "value=[1.0,2.0,3.0] dims=['test_dim'] hyper=False free=True"

    # test scripting input
    test = Param(value=np.array([1.0,2.0,3.0]))
    config.model_parameters.test = test

    # test dict input
    config.model_parameters.test = test.model_dump(exclude_none=True)
    assert config.model_parameters.test == test # type: ignore

    # test config file input
    config.model_parameters.test = io
    assert config.model_parameters.test == test  # type: ignore
    
    # test serialization
    serialized = config.model_parameters.model_dump(mode="json")
    assert serialized == {"test": io}

def test_prior():
    config = Config()

    io = "lognorm(scale=[1.0,1.0,a],s=1.5)"

    test_prior = RandomVariable(
        distribution="lognorm", 
        parameters={"scale": Expression("[1.0,1.0,a]"), "s": Expression("1.5")},
    )

    # test scripting input
    test_param = Param(value=np.array([1.0,2.0,3.0]))
    test_param.prior = test_prior

    # test dict input
    test_param.prior = test_prior.model_dump(exclude_none=True)
    assert test_param.prior == test_prior # type: ignore

    # test config file input
    test_param.prior = io
    assert test_param.prior == test_prior  # type: ignore
    
    # test serialization
    config.model_parameters.test = test_param
    serialized = test_prior.model_dump(mode="json")
    config.model_parameters.model_dump(mode="json")
    assert serialized == io

def test_parameter_array_with_prior():
    config = Config()

    io = "value=[1.0,2.0,3.0] dims=[] prior=lognorm(scale=[1.0,1.0,1.0],s=1.0) hyper=False free=True"
    test = Param(value=[1.0,2.0,3.0], prior="lognorm(scale=[1.0,1.0,1.0],s=1.0)") # type:ignore

    # test config file input
    config.model_parameters.test = io
    assert config.model_parameters.test == test  # type: ignore
    
    # test scripting input
    config.model_parameters.test = test

    # test dict input
    config.model_parameters.test = test.model_dump(exclude_none=True)
    assert config.model_parameters.test == test # type: ignore

    # test serialization
    serialized = config.model_parameters.model_dump(mode="json")
    assert serialized == {"test": io}



def test_model_parameters():
    config = Config()

    a = Param(value=1)
    b = Param(value=5, free=False)

    config.model_parameters.a = a
    config.model_parameters.b = b

    frmp = config.model_parameters.free
    fimp = config.model_parameters.fixed
    almp = config.model_parameters.all

    assert frmp == {"a": a}
    assert fimp == {"b": b}
    assert almp == {"a": a, "b":b}

def test_error_model():
    config = Config()

    io = "norm(loc=1,scale=2)"
    
    # test config file input
    test = RandomVariable(
        distribution="norm",
        parameters=dict(loc=Expression("1"),scale=Expression("2")),  # type:ignore
    )

    # test config file input
    config.error_model.test = io
    assert config.error_model.test == test  # type: ignore
    
    # test scripting input
    config.error_model.test = test

    # test dict input
    config.error_model.test = test.model_dump(exclude_none=True)
    assert config.error_model.test == test # type: ignore

    # test serialization
    serialized = config.error_model.model_dump(mode="json", exclude_none=True)
    assert serialized == {"test": io}

def test_error_model_with_obs():
    config = Config()

    io = "lognorm(scale=[1.0,1.0,1.0],s=1.0,obs=b/jnp.sqrt(2))"
    test = RandomVariable(
        distribution="lognorm",
        parameters=dict(scale=Expression("[1.0,1.0,1.0]"),s=Expression("1.0")),  # type:ignore
        obs=Expression("b/jnp.sqrt(2)")
    )

    # test config file input
    config.error_model.test = io
    assert config.error_model.test == test  # type: ignore
    
    # test scripting input
    config.error_model.test = test

    # test dict input
    config.error_model.test = test.model_dump(exclude_none=True)
    assert config.error_model.test == test # type: ignore

    # test serialization
    serialized = config.error_model.model_dump(mode="json")
    assert serialized == {"test": io}



def test_data_variables():
    config = Config()
    config.case_study.name = "lotka_volterra_case_study"
    config.case_study.scenario = "test_scenario_scripting_api"
    config.data_structure.wolves = DataVariable(dimensions=["time"], min=0)
    assert config.data_structure.data_variables == ["wolves"]
    config.save(force=True)
    
    config.data_structure = {"wolves": dict(dimensions = ["time"])} # type: ignore
    assert config.data_structure.data_variables == ["wolves"]


    config.data_structure.B = DataVariable(dimensions=["a", "b"], dimensions_evaluator=["b","a"])
    assert config.data_structure.dimdict == {"wolves": ["time"], "B": ["a", "b"]}
    assert config.data_structure.var_dim_mapper == {"wolves": [0], "B": [1,0]}



def test_commandline_api_configure():
    runner = CliRunner()
    
    args = [
        "--file=case_studies/lotka_volterra_case_study/scenarios/test_scenario_scripting_api/test_settings.cfg",
        "-o inference.numpyro.kernel=nuts",
        "-o simulation.y0 = wolves=wolves rabbits=rabbits",
        "-o inference.n_predictions=10",
        "-o model-parameters.test_parameter=",
    ]
    result = runner.invoke(configure, args)

    if result.exception is not None:
        raise result.exception


if __name__ == "__main__":
    pass


import warnings
from typing import Literal, Union, List, Optional, Dict
from functools import partial

import numpy as np
import xarray as xr
import arviz as az
import pandas as pd
from arviz.sel_utils import xarray_var_iter
from matplotlib import pyplot as plt
import matplotlib as mpl

from pymob.utils.plot_helpers import plot_hist, plot_loghist

def cluster_chains(posterior, deviation="std"):
    assert isinstance(posterior, (xr.DataArray, xr.Dataset))
    chain_means = posterior.mean(dim="draw")
    if deviation == "std":
        chain_dev = posterior.std(dim="draw")
    elif "frac:" in deviation:
        _, frac = deviation.split(":")
        chain_dev = chain_means * float(frac)
    else:
        raise ValueError("Deviation method not implemented.")    

    global cluster_id
    cluster_id = 0
    cluster = [cluster_id] * len(posterior.chain)
    unclustered_chains = posterior.chain.values

    def recurse_clusters(unclustered_chains):
        global cluster_id
        compare = unclustered_chains[0]
        new_cluster = []
        for i in unclustered_chains[1:]:
            a = chain_means.sel(chain=compare) + chain_dev.sel(chain=compare) > chain_means.sel(chain=i)
            b = chain_means.sel(chain=compare) - chain_dev.sel(chain=compare) < chain_means.sel(chain=i)
            isin_dev = (a * b).all()

            if isinstance(isin_dev, xr.Dataset):
                isin_dev = isin_dev.to_array().all()

            if not isin_dev:
                cluster[i] = cluster_id + 1
                new_cluster.append(i)

        if len(new_cluster) == 0:
            return

        cluster_id += 1
        recurse_clusters(new_cluster)
    
    recurse_clusters(unclustered_chains)

    if cluster_id > 0:
        warnings.warn(
            "The number of clusters in the InferenceData object was "+
            f"{cluster_id + 1} > 1. This indicates that not all chains/restarts "
            "Converged on the same estimate."
        )

    return cluster


def rename_extra_dims(df, extra_dim_suffix="_dim_0", new_dim="new_dim", new_coords=None):
    # TODO: COuld be used for numypro backend for fixing posterior indexes
    df_ = df.copy()
    data_vars = list(df_.data_vars.keys())

    # swap dimension names for all dims that have the suffix 
    new_dims = {}
    for dv in data_vars:
        old_dim = f"{dv}{extra_dim_suffix}"
        if df_.dims[old_dim] == 1:
            df_[dv] = df_[dv].squeeze(old_dim)
        else:
            new_dims.update({old_dim: new_dim})

    df_ = df_.swap_dims(new_dims)

    # assign coords to new dimension
    df_ = df_.assign_coords({new_dim: new_coords})
    
    # drop renamed coords
    df_ = df_.drop([f"{dv}{extra_dim_suffix}" for dv in data_vars])

    return df_



def plot_trace(idata, var_names, output):
    """
    TODO: Should be outsourced to pymob.sim.plot
    """

    if hasattr(idata, "sample_stats"):
        if "diverging" in idata["sample_stats"]:
            idata["sample_stats"]["diverging"] = idata["sample_stats"].diverging.astype(int)

    if hasattr(idata, "posterior"):
        axes = az.plot_trace(
            idata,
            var_names=var_names
        )
        fig = plt.gcf()
        fig.tight_layout()
        fig.savefig(output)

    return fig, output

def plot_pairs(idata, var_names, output):
    """
    TODO: Should be outsourced to pymob.sim.plot
    """

    if hasattr(idata, "sample_stats"):
        if "diverging" in idata["sample_stats"]:
            idata["sample_stats"]["diverging"] = idata["sample_stats"].diverging.astype(int)

    if hasattr(idata, "posterior"):
        axes = az.plot_pair(
            idata, 
            divergences=True, 
            var_names=var_names
        )
        fig = plt.gcf()
        fig.tight_layout()
        fig.savefig(output)

    return fig, output



# plot loghist
def plot_posterior_samples(posterior, col_dim=None, log=True, hist_kwargs={}):
    if log:
        hist = plot_loghist
    else:
        hist = plot_hist

    parameters = list(posterior.data_vars.keys())
    samples = posterior.stack(sample=("chain", "draw"))

    fig = plt.figure(figsize=(5, len(parameters)*2))
    fig.subplots_adjust(right=.95, top=.95, hspace=.25)

    gs = fig.add_gridspec(len(parameters), 1)

    for i, key in enumerate(parameters):
        postpar = samples[key]
        if col_dim in postpar.dims:
            col_coords = postpar[col_dim]
            gs_par = gs[i, 0].subgridspec(1, len(col_coords))
            axes = gs_par.subplots()

            for ax, coord in zip(axes, col_coords):
                hist(
                    x=postpar.sel({col_dim: coord}), 
                    name=f"${key}$ {str(coord.values)}",
                    ax=ax,
                    **hist_kwargs
                )

        else:
            gs_par = gs[i, 0].subgridspec(1, 1)
            ax = gs_par.subplots()

            hist(
                x=postpar, 
                name=f"${key}$",
                ax=ax,
                **hist_kwargs
            )


    return fig


def bic(idata: az.InferenceData):
    """calculate the BIC for az.InferenceData. The function will average over
    all samples from the markov chain
    """
    log_likelihood = idata.log_likelihood.mean(("chain", "draw")).sum().to_array().sum()
    k = idata.posterior.mean(("chain", "draw")).count().to_array().sum()

    vars = [i.split("_")[0] for i in list(idata.log_likelihood.data_vars.keys())]
    n = 0
    for v in vars:
        if v in idata.observed_data:
            n += (~idata.observed_data[v].isnull()).sum()
        elif v + "_obs" in idata.observed_data:
            n += (~idata.observed_data[v + "_obs"].isnull()).sum()
        else:
            raise IndexError(f"Variable {v} or {v+'_obs'} not in idata")

    # n = (~idata.observed_data[vars].isnull()).sum().to_array().sum()

    bic = float(k * np.log(n) - 2 * log_likelihood)
    msg = str(
        "Bayesian Information Criterion (BIC):"
        f"\nParameters: {int(k)}"
        f"\nData points: {int(n)}"
        f"\nLog-likelihood: {float(log_likelihood)}"
        f"\nBIC: {bic}"
    )

    return msg, bic

def log_lik(x):
    dims_sum = [d for d in x.dims if d not in ["chain", "draw"]]
    return x.sum(dims_sum)

def rmse(x_0: xr.DataArray, x: xr.DataArray):
    n = x.count()
    dims_sum = [d for d in x_0.dims if d not in ["chain", "draw"]]
    return np.sqrt((1 / n * np.power(x - x_0, 2).sum(dims_sum)))

def nrmse(x_0: xr.DataArray, x: xr.DataArray, mode: Literal["range", "mean", "iqr"]):
    dims_sum = [d for d in x_0.dims if d not in ["chain", "draw"]]

    if mode == "range":
        return rmse(x_0, x) / (x.max(dims_sum) - x.min(dims_sum))
    elif mode == "mean":
        return rmse(x_0, x) / x.mean(dims_sum)
    elif mode == "mean":
        return rmse(x_0, x) / x.mean(dims_sum)
    else:
        raise NotImplementedError(
            f"Mode {mode} not implemented. Use one of the following modes"
        )

        
    # normalized_obs = idata.observed_data.survival.groupby("id").map(
    #     lambda group: group / group.max(dim="time")
    # )
    # nrmse_draws = 1 / normalized_obs.mean()
    # np.sqrt(1/normalized_obs.count() * ((normalized_obs-idata.posterior_model_fits.survival ** 2).sum(dim=["id","time"]))
    # data_dict["nrmse_mean"].append(100 * float(nrmse_draws.mean().data))
    # data_dict["nrmse_higher"].append(100 * float(az.hdi(nrmse_draws,hdi_prob=0.95).sel(hdi="higher").to_dataarray()[0].data))
    # data_dict["nrmse_lower"].append(100*float(az.hdi(nrmse_draws,hdi_prob=0.95).sel(hdi="lower").to_dataarray()[0].data))

def add_cluster_coordinates(idata: az.InferenceData, deviation="std") -> az.InferenceData:
    """Clusters the chains in the posterior"""
    if "posterior" in idata.groups():
        cluster = cluster_chains(idata.posterior, deviation=deviation)
        idata = idata.assign_coords(cluster=("chain", cluster))
    return idata


def format_parameter(par, subscript_sep="_", superscript_sep="__", textwrap="\\text{}"):
    super_pos = par.find(superscript_sep)
    sub_pos = par.find(subscript_sep)
    
    scripts = sorted(zip([sub_pos, super_pos], [subscript_sep, superscript_sep]))
    scripts = [sep for pos, sep in scripts if pos > -1]


    def wrap_text(substr):

        if len(substr) == 1:
            substring_fmt = f"{substr}"
        else:
            substr = substr.replace("_", " ")
            substring_fmt = textwrap.replace("{}", "{{{}}}").format(substr)
    
        return f"{{{substring_fmt}}}"

    formatted_string = "$"
    for i, sep in enumerate(scripts):
        substr, par = par.split(sep, 1)
        substring_fmt = wrap_text(substr=substr)

        math_sep = "_" if sep == subscript_sep else "^"

        formatted_string += substring_fmt + math_sep

    formatted_string += wrap_text(par) + "$"

    return formatted_string

def round_to_sigfig(num, sig_fig=3):
    """Rounds a number to a specified number of significant figures."""
    if num == 0:
        return num
    if np.isnan(num):
        return num
    return round(num, sig_fig - int(np.floor(np.log10(abs(num)))) - 1)


def create_table(
    posterior, 
    error_metric: Literal["hdi","sd"] = "hdi", 
    vars: Dict = {}, 
    nesting_dimension: Optional[Union[List,str]] = None,
    fmt: Literal["csv", "tsv", "latex"] = "csv",
    significant_figures: int = 3,
    parameters_as_rows: bool = True,
) -> pd.DataFrame:
    """The function is not ready to deal with any nesting dimensionality
    and currently expects the 2-D case
    """
    tab = az.summary(
        posterior, var_names=list(vars.keys()), 
        fmt="xarray", kind="stats", stat_focus="mean", 
        hdi_prob=0.94
    )

    tab = tab.rename(vars)

    if nesting_dimension is None:
        stack_cols = ("metric",)
    else:
        if isinstance(nesting_dimension, str):
            nesting_dimension = [nesting_dimension]
        stack_cols = (*nesting_dimension, "metric")
        
    stack_cols = [s for s in stack_cols if s in tab.coords]


    tab = tab.apply(np.vectorize(
        partial(round_to_sigfig, sig_fig=significant_figures)
    ))


    if error_metric == "sd":
        arrays = []
        for par in vars.values():
            par_formatted = tab.sel(metric=["mean", "sd"])[par]\
                .astype(str).str\
                .join("metric", sep=" ± ")
            arrays.append(par_formatted)


        table = xr.combine_by_coords(arrays)
        table = table.assign_coords(metric="mean ± std").expand_dims("metric")
        table = table.to_dataframe().T

    elif error_metric == "hdi":
        stacked_tab = tab.sel(metric=["mean", "hdi_3%", "hdi_97%"])\
            .assign_coords(metric=["mean", "hdi 3%", "hdi 97%"])\
            .stack(col=stack_cols)
        table = stacked_tab.to_dataframe().T.drop(list(stack_cols))

    else:
        raise NotImplementedError("Must use one of 'sd' or 'hdi'")


    if fmt == "latex":
        table.columns.names = [c.replace('_',' ') for c in table.columns.names]
        table.index = [format_parameter(i) for i in list(table.index)]
        table = table.rename(
            columns={"hdi 3%": "hdi 3\\%", "hdi 97%": "hdi 97\\%"}
        )
    else: 
        pass

    if parameters_as_rows:
        return table
    else:
        return table.T


def filter_not_converged_chains(idata, deviation=1.05):
    posterior = idata.posterior
    log_likelihood = idata.log_likelihood
    log_likelihood_summed = log_likelihood.to_array("obs")
    log_likelihood_summed = log_likelihood_summed.sum(("time", "id", "obs"))

    # filter non-converged parameter estimates
    likelihood_mask = (
        # compares the mean of the summed log likelihood of a given chain
        log_likelihood_summed.mean("draw") > 
        # to the maximum of all chain means times a factor
        log_likelihood_summed.mean("draw").max() * deviation
    )
    posterior_filtered = posterior.where(likelihood_mask, drop=True)
    log_likelihood_filtered = log_likelihood.where(likelihood_mask, drop=True)

    idata = az.InferenceData(
        posterior=posterior_filtered, 
        log_likelihood=log_likelihood_filtered,
        observed_data=idata.observed_data,
    )

    return idata


def log(msg, out, newlines=1, mode="a"):
    with open(out, mode, encoding="utf-8") as f:
        print(msg, file=f, end="\n")
        for _ in range(newlines):
            print("", file=f, end="\n")


def evaluate_posterior(sim, nesting_dimension, n_samples=10_000, vars_table={}, 
                       seed=1, save=True, show=False):
    """The function is not ready to deal with any nesting dimensionality 
    and currently expects the 2-D case
    """
    rng = np.random.default_rng(seed)
    idata = sim.inferer.idata
    # idata.posterior = idata.posterior.chunk(chunks={"draw":100}).load()
    # idata.log_likelihood = idata.log_likelihood.chunk(chunks={"draw":100})
    n_subsample = min(
        int(n_samples / idata.posterior.sizes["chain"]), 
        idata.posterior.sizes["draw"]
    )

    if n_subsample < 250:
        warnings.warn(
            "The number of samples drawn from each chain for the pairplot "
            f"({n_subsample}) may be too small to be representative. "
            "Consider increasing n_samples."
        )

    subsamples = rng.choice(idata.posterior.draw, n_subsample, replace=False)
    idata.posterior = idata.posterior.sel(draw=subsamples)
    idata.log_likelihood = idata.log_likelihood.sel(draw=subsamples)

    log_likelihood_summed = idata.log_likelihood.to_array("obs")
    log_likelihood_summed = log_likelihood_summed.sum(("time", "id", "obs"))

    print(az.summary(idata.posterior))

    table = create_table(
        posterior=idata.posterior, 
        error_metric="hdi",
        vars=vars_table,
        nesting_dimension=nesting_dimension,
        fmt="latex",
    )
    table.index = [format_parameter(i) for i in list(table.index)]

    # bic 
    msg, _ = bic(idata)
    if save:
        log(table, out=f"{sim.output_path}/parameter_table.tex", mode="w")
        log(msg=msg, out=f"{sim.output_path}/bic.md", mode="w")

    if show:
        print(table)
        print(msg)

    fig_param = plot_posterior_samples(
        idata.posterior, 
        col_dim=nesting_dimension, 
        log=True,
        hist_kwargs = dict(hdi=True, bins=20)
    )
    fig_param.set_size_inches(12, 30)

    if save:
        fig_param.savefig(f"{sim.output_path}/multichain_parameter_estimates.jpg")

    if show:
        plt.show()
    else:
        plt.close()

    def plot_pairs(posterior, likelihood):
        parameters = list(posterior.data_vars.keys())

        N = len(parameters)
        parameters_ = parameters.copy()
        fig = plt.figure(figsize=(3*N, 3*(N+1)))
        gs = fig.add_gridspec(N, N+1, width_ratios=[1]*N+[0.2])
        

        i = 0
        while len(parameters_) > 0:
            par_x = parameters_.pop(0)
            hist_ax = gs[i,i].subgridspec(1, 1).subplots()
            plot_hist(
                posterior[par_x].stack(sample=("chain", "draw")), 
                ax=hist_ax, decorate=False, bins=20
            )
            hist_ax.set_title(par_x)
            for j, par_y in enumerate(parameters_, start=i+1):
                ax = gs[j,i].subgridspec(1, 1).subplots()

                scatter = ax.scatter(
                    posterior[par_x], 
                    posterior[par_y], 
                    c=likelihood, 
                    alpha=0.25,
                    s=10,
                    cmap=mpl.colormaps["plasma_r"]
                )

                if j != len(parameters)-1:
                    ax.set_xticks([])
            
                ax.set_xlabel(par_x)            
                ax.set_ylabel(par_y)

            i += 1

        # ax_colorbar = gs[:,N].subgridspec(1, 1).subplots()
        # fig.colorbar(scatter, cax=ax_colorbar)
        return fig

    for coord in idata.posterior[nesting_dimension].values:
        print("=" * len(coord))
        print(coord.capitalize())
        print("=" * len(coord))
        az.plot_trace(idata.posterior.sel({nesting_dimension:coord}))
        
        if save:
            plt.savefig(f"{sim.output_path}/multichain_pseudo_trace_{coord}.jpg")

        if show:
            plt.show()
        else:
            plt.close()

        fig = plot_pairs(
            posterior=idata.posterior.sel({nesting_dimension:coord}), 
            likelihood=log_likelihood_summed,
        )

        if save:
            fig.savefig(f"{sim.output_path}/multichain_pairs_{coord}.jpg")

        if show:
            plt.show()
        else:
            plt.close()



def plot_pair(posterior, likelihood, parameters, axes=None):

    likelihood_vars = list(likelihood.data_vars.keys())
    N = len(likelihood_vars)

    if axes is None:
        fig, axes = plt.subplots(1, N, figsize=(3+(2*(N-1)), 3), sharey=True, sharex=True)
    
    if isinstance(likelihood, xr.DataArray):
        if likelihood.name is None:
            likelihood.name = "joint_likelihood"
        likelihood = likelihood.to_dataset()
    

    par_a, par_b = parameters

    post_a = posterior[par_a]
    post_b = posterior[par_b]

    # always put the smaller parameter on the x-axis
    if post_a.ndim <= post_b.ndim:
        post_x = post_a
        post_y = post_b
    else:
        post_x = post_b.stack(sample=("chain", "draw"))
        post_y = post_a.stack(sample=("chain", "draw"))

    loglik = likelihood.stack(sample=("chain","draw"))
    extra_dims_y = [d for d in post_y.dims if d not in post_x.dims]
    norm = mpl.colors.Normalize(vmin=(likelihood).to_array().min(), vmax=(likelihood).to_array().max())
    cmap = mpl.colormaps["plasma"]

    for likvar, ax in zip(likelihood_vars, axes):
        ll = loglik[likvar]
        ax.set_title(likvar)
        ax.set_xlabel(post_x.name)
        ax.set_ylabel(post_y.name)
        for dim in extra_dims_y:
            for coord in post_y.coords[dim].values:
                if dim in ll.dims:
                    ll_coord = ll.sel({dim:coord})
                else:
                    ll_coord = ll

                scatter = ax.scatter(
                    post_x, 
                    post_y.sel({dim:coord}), 
                    c=ll_coord, 
                    alpha=0.25,
                    marker=f"${coord}$",
                    s=40,
                    cmap=cmap,
                    norm=norm
                )

                # ax.annotate()
        
    cbar = fig.colorbar(
        mappable=mpl.cm.ScalarMappable(norm=norm, cmap=cmap), 
        # if the same axis is used it is plot next to
        # the plot depending on the location option
        # ax=ax,
        # location="right",
        orientation="vertical",

        # if another axis is used, the fraction should be increased to 1.0
        ax=ax,
        fraction=0.1,
        pad=0.05,
        # aspect=20, # parameter is not so important. It thins the colorbar
        label="negative log-likelihood"
    )
    fig.tight_layout()

    # ax_colorbar = gs[:,N].subgridspec(1, 1).subplots()
    return fig


import os
import time
import numpy as np
import pytest

from pymob.sim.config import Param, DataVariable
from pymob.solvers import JaxSolver, SolverBase
from pymob.solvers.base import rect_interpolation
from tests.fixtures import (
    init_simulation_casestudy_api, 
    init_lotkavolterra_simulation_replicated,
    setup_solver
)

from pymob import SimulationBase

def test_solver_preprocessing():
    sim = init_simulation_casestudy_api()
    sim.config.model_parameters.gamma = Param(value=0.3)
    sim.config.model_parameters.delta = Param(value=0.01)
    
    # test SolverBase
    solver = setup_solver(sim, solver=SolverBase)
    y0 = sim.parse_input("y0",drop_dims=["time"])
    y0_solver = solver.preprocess_y_0(sim.validate_model_input(y0))
    np.testing.assert_equal([y.shape for y in y0_solver], [(1,1), (1,1)])

    # test jax solver
    solver = setup_solver(sim, solver=JaxSolver)
    y0 = sim.parse_input("y0",drop_dims=["time"])
    y0_solver = solver.preprocess_y_0(sim.validate_model_input(y0))
    np.testing.assert_equal([y.shape for y in y0_solver], [(1,1), (1,1)])

def test_solver_preprocessing_replicated():
    sim = init_lotkavolterra_simulation_replicated()

    # test SolverBase
    solver = setup_solver(sim, solver=SolverBase)
    y0 = sim.parse_input("y0",drop_dims=["time"])
    y0_solver = solver.preprocess_y_0(sim.validate_model_input(y0))
    np.testing.assert_equal([y.shape for y in y0_solver], [(2,1), (2,1)])
    np.testing.assert_equal(y0_solver, [np.array([[9],[5]]), np.array([[40], [50]])])

    sim.config.simulation.batch_dimension = "idx" # this is not existing
    solver = setup_solver(sim, solver=JaxSolver)
    y0 = sim.parse_input("y0",drop_dims=["time"])
    y0_solver = solver.preprocess_y_0(sim.validate_model_input(y0))
    np.testing.assert_equal([y.shape for y in y0_solver], [(1,2), (1,2)])
    np.testing.assert_equal(y0_solver, [np.array([[9, 5]]), np.array([[40, 50]])])
    
    sim.config.simulation.batch_dimension = "id" # This exists!
    solver = setup_solver(sim, solver=JaxSolver)
    y0 = sim.parse_input("y0",drop_dims=["time"])
    y0_solver = solver.preprocess_y_0(sim.validate_model_input(y0))
    np.testing.assert_equal([y.shape for y in y0_solver], [(2,1), (2,1)])
    np.testing.assert_equal(y0_solver, [np.array([[9],[5]]), np.array([[40], [50]])])
    
    # test parameter processing
    theta = sim.model_parameter_dict
    theta_solver_ode, theta_solver_pp = solver.preprocess_parameters(theta)
    np.testing.assert_equal(
        [t.shape for t in theta_solver_ode], 
        [(2,1), (2,1), (2,1), (2,1)]
    )


def test_solver_preprocessing_complex_parameters():
    sim = init_lotkavolterra_simulation_replicated()

    # step 1 of mischief ;) add a parameter with more than 1 dimension
    # without declaring these coordinates. By using the same number
    # of parameters as the batch_dimension no error is thrown.
    # This is an expected behavior, but will issue a warning if declared
    # without dimensions
    sim.config.model_parameters.gamma.value = [0,1]
    sim.config.model_parameters.gamma.dims = ("id",)

    # test SolverBase
    solver = setup_solver(sim, solver=SolverBase)

    # test parameter processing
    theta = sim.model_parameter_dict
    theta_solver_ode, theta_solver_pp = solver.preprocess_parameters(theta)

    # step 2 of mischief ;) this is not allowed and should raise an exception
    sim.config.model_parameters.gamma.value = [0,1,2]
    sim.config.model_parameters.gamma.dims = ()

    # test parameter processing
    solver = setup_solver(sim, solver=SolverBase)
    theta = sim.model_parameter_dict

    try:
        theta_solver_ode, theta_solver_pp = solver.preprocess_parameters(theta)
        raise AssertionError(
            "This behavior should throw an exception. Multidimensional"+
            "Array values were specified without explicitly defining dimensions"
        )
    except ValueError:
        pass
    
    # step 3: fixing mischief conducted
    # there should be a quick fix for this
    # TODO: Add a dummy coordinate for the parameter dimension.
    sim.config.model_parameters.gamma.dims = ("test_dim",)
    
    try:
        sim.dimension_coords
        raise AssertionError(
            "This behavior should throw an exception. Parameter dimension"+
            "was specified without explicitly defining coordinates for that "+
            "dimension."
        )
    except KeyError:
        pass
    

    # step 4: try to fix but should not work because dimension size does not match 
    # value shape
    sim.coordinates["test_dim"] = ["a", "b"]

    sim.dimension_coords
    sim.dimension_sizes

    solver = setup_solver(sim, solver=SolverBase)
    theta = sim.model_parameter_dict

    try:
        theta_solver_ode, theta_solver_pp = solver.preprocess_parameters(theta)
        raise AssertionError(
            "This behavior should throw an exception. Multidimensional"+
            "Array values were specified without explicitly defining dimensions"
        )
    except ValueError:
        pass
    
    
    # step 5: Correct the coordinate shape
    sim.coordinates["test_dim"] = ["a", "b", "c"]

    solver = setup_solver(sim, solver=SolverBase)
    theta = sim.model_parameter_dict
    theta_solver_ode, theta_solver_pp = solver.preprocess_parameters(theta)

    np.testing.assert_equal(
        [t.shape for t in theta_solver_ode], 
        [(2,1), (2,1), (2,3), (2,1)]
    )

def test_solver_dimensional_order():
    sim = init_lotkavolterra_simulation_replicated()
    theta = sim.model_parameter_dict
    sim.solver = JaxSolver

    sim.data_structure_and_dimensionality
    sim.dispatch_constructor()
    e = sim.dispatch(theta)
    e()
    res_id_time = e.results

    # reorder dimensions of the data variables to see if they can be processed
    sim.config.data_structure.wolves.dimensions = ["time", "id"] # type: ignore
    sim.config.data_structure.rabbits.dimensions = ["time", "id"] # type: ignore

    sim.dispatch_constructor()
    e = sim.dispatch(theta)
    e()
    res_time_id = e.results

    np.testing.assert_equal(
        (res_id_time.to_array() - res_time_id.to_array()).values, 0
    )




def test_benchmark_time():
    sim = init_simulation_casestudy_api()

    cpu_time_start = time.process_time()
    sim.benchmark(n=100)
    cpu_time_stop = time.process_time()

    t = cpu_time_stop - cpu_time_start

    if t > 2:
        raise AssertionError(f"Benchmarking took too long: {t}s. Expected t < 2s")


def test_benchmark_jaxsolver():
    sim = init_simulation_casestudy_api()

    # dispatch is constructed in `init_simulation_case_study`
    e = sim.dispatch({})
    e()
    a = e.results

    # construct the dispatch again with a different solver
    sim.solver = JaxSolver
    from diffrax import Dopri5
    sim.dispatch_constructor(diffrax_solver=Dopri5, rtol=1e-6, atol=1e-7)
    e = sim.dispatch({})
    e()
    b = e.results

    np.testing.assert_allclose(a.to_array(), b.to_array(), atol=1e-3, rtol=1e-5)

    cpu_time_start = time.process_time()
    sim.benchmark(n=100)
    cpu_time_stop = time.process_time()

    t = cpu_time_stop - cpu_time_start

    if t > 1:
        raise AssertionError(f"Benchmarking took too long: {t}s. Expected t < 1s")


def test_rect_interpolation():
    # TODO: Use another test for making sure, the interpolation works. This
    # is not the right place. A mini Simulation using interpolated data would
    # be great.
    sim: SimulationBase
    
    # TODO: Use another test for making sure, the interpolation works. This
    # is not the right place. A mini Simulation using interpolated data would
    # be great.
    sim = None
    pytest.skip()
    sim.use_jax_solver() # type: ignore

    # x input is defined on the interval [0,179]
    x_in = sim.parse_input(input="x_in", reference_data=sim.observations, drop_dims=[])
    # rect_interpolation adds duplicates the last y_179 for x_180
    x_in = rect_interpolation(x_in=x_in, x_dim="time")
    sim.model_parameters["x_in"] = x_in

    # Interpolations in diffrax now jump each discontinuity until the last time
    # that is evaluated by the solver. This is not jumped, so that it can be 
    # retrieved by SaveAt
    # https://github.com/patrick-kidger/diffrax/issues/58
    # It seems like this is the intended behavior of diffrax, 

    # run the simulation until exactly the last time, which is a discontinuity
    # this works and will not return a discontinuity, because jump_ts in the
    # PIDController of diffrax, is told that it shoud jump all ts that are 
    # smaller than x_stop
    sim.coordinates["time"] = np.linspace(0, sim.t_max - 1, 1000) #type: ignore
    sim.dispatch_constructor(max_steps=1e5, throw_exception=True, pcoeff=0.0, icoeff=0.25)
    e = sim.dispatch(theta={})
    e()

    # assert that the interpolation produces no infinity values 
    np.testing.assert_array_equal(
        (e.results == np.inf).sum().to_array().values, 
        np.array([0, 0, 0, 0, 0])
    )

    # test if the simulation also works with the normal time vector
    sim.reset_coordinate(dim="time")
    sim.dispatch_constructor(max_steps=1e5, throw_exception=True, pcoeff=0.0, icoeff=0.25)
    e = sim.dispatch(theta={})
    e()

    # assert that the interpolation produces no infinity values 
    np.testing.assert_array_equal(
        (e.results == np.inf).sum().to_array().values, 
        np.array([0, 0, 0, 0, 0])
    )

    # run the simulaton until the added interpolation provided by rect_interpolation
    # until t=180
    sim.coordinates["time"] = np.linspace(0, sim.t_max, 1000) # type: ignore
    sim.dispatch_constructor(max_steps=1e5, throw_exception=True, pcoeff=0.0, icoeff=0.25)
    e = sim.dispatch(theta={})
    e()

    # assert that the interpolation works without problems if the last specified
    # value is not a discontinuity
    np.testing.assert_array_equal(
        (e.results == np.inf).sum().to_array().values, 
        np.array([0, 0, 0, 0, 0])
    )

    # run the simulaton until the added interpolation provided by rect_interpolation
    # until t=500
    # This behavior is correctly caught, by pymob, before it can ocurr, because
    # an interpolation over the provided ts, and ys is not possible. And would
    # result in a difficult to diagnose max_steps error
    try:
        sim.coordinates["time"] = np.linspace(0, sim.t_max + 0.01, 1000) # type: ignore
        sim.dispatch_constructor(max_steps=1e5, throw_exception=True, pcoeff=0.0, icoeff=0.25)
        threw_error = False
    except AssertionError:
        threw_error = True

    if not threw_error: 
        AssertionError(
            "The solver should fail if interpolation is attempted "
            "to be done further than the intended range."        
        )

def test_no_interpolation():
    # TODO: Use another test for making sure, the interpolation works. This
    # is not the right place
    sim: SimulationBase
    # TODO: Use another test for making sure, the interpolation works. This
    # is not the right place. A mini Simulation using interpolated data would
    # be great.
    sim = None
    pytest.skip()

    sim.use_jax_solver() # type: ignore
    
    # x input is defined on the interval [0,179]
    x_in = sim.parse_input(input="x_in", reference_data=sim.observations, drop_dims=[])
    sim.model_parameters["x_in"] = x_in

    # run the simulaton until the provided x_input until t=179
    sim.coordinates["time"] = np.linspace(0, sim.t_max - 1, 1000) # type: ignore

    # without interpolation, the time needs to be 
    sim.dispatch_constructor(max_steps=1e6, throw_exception=True, pcoeff=0.0, icoeff=0.25)
    e = sim.dispatch(theta={})
    e()

    # assert that the interpolation works without problems if the last specified
    # value is not a discontinuity
    np.testing.assert_array_equal(
        (e.results == np.inf).sum().to_array().values, 
        np.array([0, 0, 0, 0, 0])
    )


if __name__ == "__main__":
    import sys
    import os
    sys.path.append(os.getcwd())
    # test_solver_dimensional_order()
    # test_solver_preprocessing_replicated()
    # test_benchmark_jaxsolver()


import pytest
import numpy as np
from click.testing import CliRunner
from matplotlib import pyplot as plt
from pymob.solvers.diffrax import JaxSolver
from pymob.inference.numpyro_backend import NumpyroBackend
from pymob.sim.parameters import Param

from tests.fixtures import (
    init_simulation_casestudy_api, 
    init_lotka_volterra_case_study_hierarchical_from_settings,
)

def test_diffrax_exception():
    # with proper scripting API define JAX model here or import from fixtures
    sim = init_simulation_casestudy_api("test_scenario")

    # diffrax returns infinity for all computed values after which the solver 
    # breaks due to raching maximum number of steps. 
    # This function calculates the number of inf values
    n_inf = lambda x: (x.results.to_array() == np.inf).values.sum() / len(x.data_variables)

    ub_alpha = 5.0  # alpha values above do not yield reasonable fits for beta = 0.02
    alpha = np.logspace(-2, 3, 20)  # we sample alpha from 0.01-100

    badness = []
    for a in alpha:
        eva = sim.dispatch({"alpha": a, "beta": 0.02})
        eva()

        badness.append(n_inf(eva))

    # the tests make sure that parameters within feasible bounds result in simulation
    # results without inf values and do contain inf values when parameters above
    # feasible bounds are sampled.
    badness_for_feasible_alpha = np.array(badness)[np.where(alpha < ub_alpha)[0]]
    assert sum(badness_for_feasible_alpha) == 0

    badness_for_infeasible_alpha = np.array(badness)[np.where(alpha >= ub_alpha)[0]]
    assert sum(badness_for_infeasible_alpha) > 0


def test_convergence_user_defined_probability_model():
    sim = init_simulation_casestudy_api("test_scenario")

    sim.config.data_structure.wolves.observed = False
    sim.config.data_structure.rabbits.observed = False

    sim.config.inference_numpyro.kernel = "nuts"
    sim.config.inference_numpyro.user_defined_probability_model = "parameter_only_model"
    sim.config.inference_numpyro.user_defined_preprocessing = "dummy_preprocessing"

    sim.set_inferer(backend="numpyro")
    sim.inferer.run()
    
    posterior_median = sim.inferer.idata.posterior.median( # type: ignore
        ("chain", "draw"))[["beta", "alpha"]] 
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_median.to_dataarray().values,
        np.array([0.02, 0.5]),
        rtol=1e-1, atol=1e-3
    )


def test_convergence_nuts_kernel():
    sim = init_simulation_casestudy_api("test_scenario")

    sim.config.inference_numpyro.kernel = "nuts"
    sim.set_inferer(backend="numpyro")
    sim.inferer.run()
    
    posterior_mean = sim.inferer.idata.posterior.mean( # type: ignore
        ("chain", "draw"))[sim.model_parameter_names]
    true_parameters = sim.model_parameter_dict
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_mean.to_dataarray().values,
        np.array(list(true_parameters.values())),
        rtol=1e-2, atol=1e-3
    )

def test_convergence_nuts_kernel_jaxsolver():
    sim = init_simulation_casestudy_api("test_scenario")
    sim.solver = JaxSolver
    # TODO can the lotka volterra case study be run with the Jax solver
    # without throw_exception = False
    sim.dispatch_constructor(throw_exception=False)
    
    sim.config.inference_numpyro.kernel = "nuts"
    sim.set_inferer(backend="numpyro")
    sim.inferer.run()
    
    posterior_mean = sim.inferer.idata.posterior.mean( # type: ignore
        ("chain", "draw"))[sim.model_parameter_names]
    true_parameters = sim.model_parameter_dict
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_mean.to_dataarray().values,
        np.array(list(true_parameters.values())),
        rtol=1e-2, atol=1e-3
    )


def test_convergence_svi_kernel_jaxsolver_truncated_alpha():
    sim = init_simulation_casestudy_api("test_scenario_truncated_alpha")
    sim.solver = JaxSolver
    # TODO can the lotka volterra case study be run with the Jax solver
    # without throw_exception = False
    sim.dispatch_constructor(throw_exception=False)
    
    sim.config.inference_numpyro.gaussian_base_distribution = True

    sim.config.inference_numpyro.kernel = "svi"
    sim.config.inference_numpyro.svi_iterations = 10_000
    sim.config.inference_numpyro.svi_learning_rate = 0.01
    sim.set_inferer(backend="numpyro")
    sim.inferer.run()
    
    posterior_mean = sim.inferer.idata.posterior.mean( # type: ignore
        ("chain", "draw"))[sim.model_parameter_names]
    true_parameters = sim.model_parameter_dict
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_mean.to_dataarray().values,
        np.array(list(true_parameters.values())),
        rtol=1e-2, atol=1e-3
    )

    sim.inferer.plot_diagnostics()

    # TODO: Test traceplot

def test_convergence_svi_kernel():
    sim = init_simulation_casestudy_api("test_scenario")

    sim.config.inference_numpyro.kernel = "svi"
    sim.config.inference_numpyro.svi_iterations = 10_000
    sim.config.inference_numpyro.svi_learning_rate = 0.01
    # this samples the model with standard normal distributions
    # and rescales them according to the transformations of the specified 
    # parameter distributions to the normal
    sim.config.inference_numpyro.gaussian_base_distribution = True

    sim.set_inferer(backend="numpyro")
    sim.inferer.run()
    sim.inferer.idata.posterior_predictive # type: ignore

    posterior_mean = sim.inferer.idata.posterior.mean( # type: ignore
        ("chain", "draw"))[sim.model_parameter_names]
    true_parameters = sim.model_parameter_dict
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_mean.to_dataarray().values,
        np.array(list(true_parameters.values())),
        rtol=1e-2, atol=1e-3
    )


    # posterior predictions
    fig, axes = plt.subplots(2,1, sharex=True)
    for data_var, ax in zip(sim.config.data_structure.data_variables, axes):
        ax = sim.inferer.plot_posterior_predictions( # type: ignore
            data_variable=data_var, 
            prediction_data_variable=data_var,
            x_dim="time",
            ax=ax
        )


def test_convergence_map_kernel():
    sim = init_simulation_casestudy_api("test_scenario")

    sim.config.inference_numpyro.kernel = "map"
    sim.config.inference_numpyro.svi_iterations = 2000
    sim.config.inference_numpyro.svi_learning_rate = 0.01
    # this samples the model with standard normal distributions
    # and rescales them according to the transformations of the specified 
    # parameter distributions to the normal
    sim.config.inference_numpyro.gaussian_base_distribution = True

    sim.set_inferer(backend="numpyro")
    sim.inferer.run()
    sim.inferer.idata.posterior_predictive # type: ignore

    posterior_mean = sim.inferer.idata.posterior.mean( # type: ignore
        ("chain", "draw"))[sim.model_parameter_names]
    true_parameters = sim.model_parameter_dict
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_mean.to_dataarray().values,
        np.array(list(true_parameters.values())),
        rtol=1e-2, atol=1e-3
    )


def test_convergence_sa_kernel():
    pytest.skip()
    sim = init_simulation_casestudy_api("test_scenario")

    sim.config.inference_numpyro.kernel = "sa"
    sim.config.inference_numpyro.init_strategy = "init_to_sample"
    sim.config.inference_numpyro.warmup = 2000
    sim.config.inference_numpyro.draws = 1000
    sim.config.inference_numpyro.sa_adapt_state_size = 10

    sim.set_inferer(backend="numpyro")
    sim.inferer.run()
    
    posterior_mean = sim.inferer.idata.posterior.mean( # type: ignore
        ("chain", "draw"))[sim.model_parameter_names]
    true_parameters = sim.model_parameter_dict
    
    # tests if true parameters are close to recovered parameters from simulated
    # data
    np.testing.assert_allclose(
        posterior_mean.to_dataarray().values,
        np.array(list(true_parameters.values())),
        rtol=1e-2, atol=1e-3
    )



    # posterior predictions
    for data_var in sim.config.data_structure.data_variables:
        ax = sim.inferer.plot_posterior_predictions( # type: ignore
            data_variable=data_var, 
            prediction_data_variable=data_var,
            x_dim="time"
        )


def test_convergence_hierarchical_lotka_volterra():
    sim = init_lotka_volterra_case_study_hierarchical_from_settings()

    sim.config.model_parameters.alpha_species_hyper = Param(
        prior="halfnorm(scale=5)", dims=('rabbit_species',) , # type: ignore
        hyper=True, free=True
    )
    # sim.config.model_parameters.alpha_species_sigma_hyper = Param(
    #     prior="halfnorm(scale=5)", dims=('rabbit_species',) , # type: ignore
    #     hyper=True, free=True
    # )
    sim.config.model_parameters.alpha_sigma = Param(
        prior="halfnorm(scale=1)", hyper=True, free=True # type:ignore
    )
    sim.config.model_parameters.alpha_species.prior="norm(loc=[[alpha_species_hyper[0]],[alpha_species_hyper[1]]],scale=0.1)" # type: ignore
    sim.config.model_parameters.alpha.prior="lognorm(s=alpha_sigma,scale=alpha_species[rabbit_species_index, experiment_index])" # type: ignore
    sim.config.model_parameters.beta.prior="lognorm(s=1,scale=1)" # type: ignore
    # sim.config.model_parameters.sigma = Param(value=1.0, prior="halfnorm(scale=5)", free=True) # type: ignore
    
    # this reorders the parameters, beginning with alpha_species_hyper
    sim.config.model_parameters.reorder(["alpha_species_hyper", "alpha_sigma"])

    sim.config.error_model.rabbits = "norm(loc=0, scale=1, obs=(obs-rabbits)/jnp.sqrt(rabbits+1e-6),obs_inv=res*jnp.sqrt(rabbits+1e-6)+rabbits)"
    sim.config.error_model.wolves = "norm(loc=0, scale=1, obs=(obs-wolves)/jnp.sqrt(wolves+1e-6),obs_inv=res*jnp.sqrt(wolves+1e-6)+wolves)"


    # using SVI, because it is much faster than NUTS. 
    sim.config.inference_numpyro.kernel = "svi"
    sim.config.inference_numpyro.svi_iterations = 2_000
    sim.config.inference_numpyro.svi_learning_rate = 0.01
    sim.config.inference_numpyro.gaussian_base_distribution = True
    sim.config.jaxsolver.max_steps = 1e5
    sim.config.jaxsolver.throw_exception = False
    sim.config.inference_numpyro.init_strategy = "init_to_median"
    sim.solver = JaxSolver
    sim.dispatch_constructor()
    sim.set_inferer("numpyro")

    sim.config.case_study.scenario = "lotka_volterra_hierarchical_presimulated_v1"
    sim.config.create_directory("scenario", force=True)
    sim.config.save(force=True)
    sim.inferer.run()

    # this works
    np.testing.assert_allclose(
        sim.inferer.idata.posterior.beta.mean(("chain", "draw")), 
        sim.config.model_parameters.beta.value,
        atol=0.0005,
        rtol=0.025
    )

    # TODO: CUrrently this is not very accurate. But it is a sufficient test
    # that hierarchical models can be fitted
    # correct
    np.testing.assert_allclose(
        sim.inferer.idata.posterior.alpha_species_hyper.mean(("chain", "draw")), 
        (1, 3),
        atol=0.1,
        rtol=0.2
    )



def test_hierarchical_lotka_volterra_user_defined_prob_model():
    pytest.skip()    
    sim = init_lotka_volterra_case_study_hierarchical_from_settings()
    sim.config.inference_numpyro.user_defined_probability_model = "hierarchical_lotka_volterra"

    sim.solver = JaxSolver
    sim.config.inference_numpyro.kernel = "svi"
    sim.config.inference_numpyro.svi_iterations = 2_000
    sim.config.inference_numpyro.svi_learning_rate = 0.005
    sim.config.jaxsolver.max_steps = 1e5
    sim.config.jaxsolver.throw_exception = False
    sim.config.inference_numpyro.init_strategy = "init_to_uniform"
    sim.dispatch_constructor()
    sim.set_inferer("numpyro")

    # TODO: Test this model with SVI
    sim.inferer.run()

    np.testing.assert_almost_equal(
        sim.inferer.idata.posterior.beta, 
        sim.config.model_parameters.beta.value
    )
    np.testing.assert_almost_equal(
        sim.inferer.idata.posterior.alpha_species, 
        sim.config.model_parameters.alpha_species.value
    )


def test_commandline_api_infer():
    from pymob.infer import main
    runner = CliRunner()
    
    args = "--case_study=lotka_volterra_case_study "+\
        "--scenario=test_scenario "+\
        "--inference_backend=numpyro"
    result = runner.invoke(main, args.split(" "))

    if result.exception is not None:
        raise result.exception


if __name__ == "__main__":
    pass


from functools import partial
from typing import Dict, Tuple, Union

import numpy as np
from numpy.random import Generator, PCG64
from scipy.stats._distn_infrastructure import rv_continuous, rv_discrete

from pymob.sim.parameters import scipy_to_scipy
from pymob.inference.base import Distribution, Errorfunction, InferenceBackend
from pymob.simulation import SimulationBase
from pymob.utils.config import lookup

class ScipyDistribution(Distribution):
    distribution_map: Dict[str,Tuple[Union[rv_continuous,rv_discrete],Dict[str,str]]] = scipy_to_scipy
    parameter_converter = staticmethod(lambda x: np.array(x))
    
    @property
    def dist_name(self) -> str:
        return self.distribution.name
    

    
class ScipyBackend(InferenceBackend):
    _distribution = ScipyDistribution
    distribution: Union[rv_continuous,rv_discrete]

    def __init__(self, simulation: SimulationBase) -> None:
        super().__init__(simulation)
        self.random_state = Generator(PCG64(self.config.simulation.seed))

    def parse_deterministic_model(self):
        pass

    def parse_probabilistic_model(self):
        pass

    def posterior_predictions(self):
        pass

    def prior_predictions(self):
        pass

    def sample_distribution(self):
        prior_samples = {}

        observations = self.simulation.observations
        # prior is added here, so it is updated 
        context=[prior_samples,self.indices,observations]
        for name, prior in self.prior.items():

            dist = prior.construct(context=context)
            sample = dist.rvs(size=prior.shape, random_state=self.random_state)
            dist.rvs()

            prior_samples.update({name:sample})

        return prior_samples
    
    def create_log_likelihood(self) -> Tuple[Errorfunction,Errorfunction]:
        # TODO: define
        return 


import ast
import warnings
import inspect
from typing import Optional, List, Dict, Tuple, Any, Union
from typing_extensions import Annotated

import scipy
import numpy as np
import xarray as xr
from pydantic import BaseModel, ConfigDict, model_serializer, field_validator, model_validator
from pydantic.functional_validators import BeforeValidator
from numpydantic import NDArray, Shape

NumericArray = Union[
    NDArray[Shape["*, ..."], float],  # noqa: F722
    NDArray[Shape["*, ..."], int],  # noqa: F722
] 

class Expression:
    """Random variables are context dependent. They may be dependent on other
    Variables, or datasets. In the config they represent an abstract structure,
    so they remain unevaluated expressions that must follow python syntax.
    Once, the context is available, the expressions can be evaluated by
    `Expression.evaluate(context={...})`.
    """
    def __init__(self, expression: Union[str, ast.Expression]):
        if isinstance(expression, str):
            self.expression = ast.parse(expression, mode="eval")
        elif isinstance(expression, ast.Expression):
            self.expression = expression
        else:
            self.expression = ast.Expression(expression)

        finder = UndefinedNameFinder()
        self.undefined_args = finder.find_undefined_names(self.expression)
        self.compiled_expression = compile(self.expression, '', mode='eval')

    def __repr__(self):
        return str(self)

    def __str__(self) -> str:
        return ast.unparse(self.expression)
    
    def evaluate(self, context: Dict = {}) -> Any:
        try:
            val = eval(self.compiled_expression, context)
            return val
        except NameError as err:
            raise NameError(
                f"{err}. Have you forgotten to pass a context?"
            )

class UndefinedNameFinder(ast.NodeVisitor):
    # powered by ChatGPT
    def __init__(self):
        self.defined_names = set()
        self.undefined_names = set()

    def visit_FunctionDef(self, node):
        # Function arguments are considered defined within the function
        for arg in node.args.args:
            self.defined_names.add(arg.arg)
        self.generic_visit(node)

    def visit_Name(self, node):
        if isinstance(node.ctx, ast.Store):
            # If the name is being assigned to, add it to the defined names
            self.defined_names.add(node.id)
        elif isinstance(node.ctx, ast.Load):
            # If the name is being used, check if it's defined
            if node.id not in self.defined_names:
                self.undefined_names.add(node.id)

    def find_undefined_names(self, expr):
        tree = ast.parse(expr, mode='exec')
        self.visit(tree)
        return self.undefined_names


class RandomVariable(BaseModel):
    """Basic infrastructure to parse priors into their components so that 
    they can be more easily parsed to other backends.
    Parsing to distributions needs more context:
    - other parameters (for hierarchical parameter structures)
    - data structure and coordinates to identify dimension sizes
    """
    model_config = ConfigDict(
        arbitrary_types_allowed=True, 
        validate_assignment=True,
        extra="forbid"
    )

    distribution: str
    parameters: Dict[str, Expression]
    obs: Optional[Expression] = None
    obs_inv: Optional[Expression] = None

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, RandomVariable):
            raise NotImplementedError("Only compare to Prior instance")
        
        return bool(
            self.distribution == other.distribution and
            self.parameters.keys() == other.parameters.keys() and
            all([str(self.parameters[k]) == str(other.parameters[k]) 
                for k in self.parameters.keys()])
        )

    @model_serializer(when_used="json", mode="plain")
    def model_ser(self) -> str:
        distribution = self.distribution
        parameters = dict_to_string(self.parameters, jstr=",")
        if self.obs is None:
            pass
        else:
            obs = str(self.obs).replace(" ","")
            parameters = parameters + f",obs={obs}"
        if self.obs_inv is None:
            pass
        else:
            obs_inv = str(self.obs_inv).replace(" ","")
            parameters = parameters + f",obs_inv={obs_inv}"
        return f"{distribution}({parameters})"

    @field_validator("distribution", mode="after")
    def check_distribution(cls, new_value, info, **kwargs):
        if new_value not in scipy_to_scipy:
            warnings.warn(
                f"The distribution '{new_value}' is not part of the scipy "+
                "distributions implemented in pymob. "+
                "This can lead to inconsistent behavior. "+
                "It is recommended to use the scipy distribution protocol "+
                "where possible. "+
                "https://docs.scipy.org/doc/scipy/reference/stats.html "+
                "It may also be possible that your distribution has not yet "+
                "been introduced into the pymob package "
            )
        return new_value.lower()

    @field_validator("parameters", mode="after")
    def check_parameters(cls, new_value, info, **kwargs):
        distribution = info.data.get("distribution")
        dist = scipy_to_scipy.get(distribution, (None, ))[0]
        if dist is None:
            return new_value
        
        if distribution == "deterministic":
            return new_value
        
        dist_args = () if dist.shapes is None else dist.shapes
        dist_params = ("loc", "scale", *dist_args)
        unmatched_params = [k for k in new_value.keys() if k not in dist_params]
        if len(unmatched_params) > 0:
            warnings.warn(
                f"The parameters '{unmatched_params}' did not follow the scipy "+
                "protocol. "+
                "This can lead to inconsistent behavior. "+
                "It is recommended to use the scipy distribution protocol "+
                "where possible. "+
                "https://docs.scipy.org/doc/scipy/reference/stats.html"
            )
        return new_value

def string_to_prior_dict(prior_str: str):
    # Step 1: Parse the string to extract the function name and its arguments.
    node = ast.parse(source=prior_str, mode='eval')
    
    if not isinstance(node, ast.Expression):
        raise ValueError("The input must be a valid Python expression.")
    
    # Extract the function name (e.g., 'ZeroSumNormal')
    func_name = node.body.func.id
    
    # Extract the keyword arguments
    kwargs = {}
    for kw in node.body.keywords:
        key = kw.arg  # Argument name (e.g., 'loc')
        # if this is a valid python expression it can be compiled, but
        # evaluated, it can only be if the respective arguments are present
        # value = compile(ast.Expression(kw.value), '', mode='eval')
        # value = eval(compile(ast.Expression(kw.value), '', mode='eval'), {"wolves": 2, "EPS": 2})  # Evaluate the value part
        value = Expression(kw.value)
        kwargs[key] = value
    
    obs = kwargs.pop("obs", None)
    obs_inv = kwargs.pop("obs_inv", None)

    # Step 3: Return the symbolic expression and the argument dictionary
    return {
        "distribution": func_name, 
        "parameters": kwargs, 
        "obs": obs,
        "obs_inv": obs_inv
    }

def to_rv(option: Union[str,RandomVariable,Dict]) -> RandomVariable:
    if isinstance(option, RandomVariable):
        return option
    elif isinstance(option, Dict):
        prior_dict = option
    else:
        prior_dict = string_to_prior_dict(option)

    return RandomVariable.model_validate(prior_dict, strict=False)


def dict_to_string(dct: Dict, jstr=" "):
    string_items = []
    for k, v in dct.items():
        if isinstance(v, np.ndarray):
            v = v.tolist()

        expr = f"{k}={v}".replace(" ", "")
        string_items.append(expr)

    return jstr.join(string_items)



OptionRV = Annotated[
    RandomVariable, 
    BeforeValidator(to_rv), 
]

class Param(BaseModel):
    """This class serves as a Basic model for declaring parameters
    Including a distribution with optional depdendencies

    Parameters
    ----------

    value : float|NumericArray
        The parameter value. If it is not a 0-d array, float or 1-d array of 
        length one, it should be accompanied by a dimension for each axis in 
        the array. The array coordinates must be specified in the observation
        coordinates or in index coordinates

    dims : Tuple[str, ...]
        Dimensions of the parameter. If the batch dimension is not specified
        here, it will be automatically added in the Evaluator 
        (dispatch_constructor). If dims are specified here, they should be
        present in:
        1) dimension in sim.config.data_structure and sim.coordinates  
        2) dimension in sim.config.parameters (so here) and sim.coordinates  
        3) dimension in sim.indices 

    prior : Optional[RandomVariable]
        This is a string or pymob.sim.parameters.RandomVariable that specifies 
        the prior. Strings are automatically parsed to RandomVariables if the 
        syntax is correct. The prior should follow the specification of 
        scipy.stats
    """
    model_config = ConfigDict(
        arbitrary_types_allowed=True, 
        validate_assignment=True, 
        extra="forbid"
    )
    name: Optional[str] = None
    value: float|NumericArray = 0.0
    dims: Tuple[str, ...] = ()
    prior: Optional[OptionRV] = None
    min: Optional[float|NumericArray] = None
    max: Optional[float|NumericArray] = None
    step: Optional[float|NumericArray] = None
    hyper: bool = False
    free: bool = True

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Param):
            raise NotImplementedError("Only compare to Param instance")
        
        return bool(
            self.name == other.name and
            self.dims == self.dims and
            np.all(self.value == other.value) and
            np.all(self.min == other.min) and
            np.all(self.max == other.max) and
            np.all(self.step == other.step) and
            self.prior == other.prior and
            self.hyper == other.hyper and
            self.free == other.free
        )
  
    @model_validator(mode="after")
    def post_update(self):
        shape_value = np.array(self.value, ndmin=1).shape
        if max(shape_value) > 1 and len(self.dims) == 0:
            warnings.warn(
                f"Declaring parameter values of a shapes {shape_value} > 1"+
                "without specifiying Param(..., dims=(...)) is dangerous."+
                "If the dimension should represent only the batch dimension (e.g.)"+
                "the replicate dimension, you can simply add it here."+
                "If the dimension is part of the model, you should absolutely"+
                "specify it along with one of the following options: "+
                "1) dimension in sim.config.data_structure and sim.coordinates "+
                "2) dimension in sim.config.parameters and sim.coordinates "+
                "3) dimension in sim.indices "+
                "an index or a coordinate and DataVariable."+
                "If this dimension is not somehow part of the datastructure of the"+
                "simulation consider if you really need it."
            )

        return self
    
    def to_xarray(self, coordinates):
        """Converts the parameter to an xarray based on the dimensional 
        structure and adds additional info (prior, starting values, ...) 
        as metadata.
        """

        coords = {d:list(coordinates[d]) for d in self.dims}
        shape = tuple([len(c) for _, c in coords.items()])
        value = np.broadcast_to(np.array(self.value), shape)
        arr = xr.DataArray(value, dims=self.dims, coords=coords)

        for key, value in self.model_dump().items():
            if value is None: 
                continue

            if key == "prior":
                arr.attrs["prior"] = self.prior.model_dump_json().strip('"') #type:ignore
            
            elif key in ["value", "dims"]:
                continue
            
            else:
                arr.attrs[key] = value

        return arr


scipy_to_scipy = {
    # Continuous Distributions
    "norm": (scipy.stats.norm, {}),
    "normal": (scipy.stats.norm, {}),
    "expon": (scipy.stats.expon, {}),
    "exponential": (scipy.stats.expon, {}),
    "uniform": (scipy.stats.uniform, {}),
    "beta": (scipy.stats.beta, {}),
    "gamma": (scipy.stats.gamma, {}),
    "lognorm": (scipy.stats.lognorm, {}),
    "lognormal": (scipy.stats.lognorm, {}),
    "chi2": (scipy.stats.chi2, {}),
    "pareto": (scipy.stats.pareto, {}),
    "t": (scipy.stats.t, {}),
    "cauchy": (scipy.stats.cauchy, {}),
    "weibull_min": (scipy.stats.weibull_min, {}),
    "weibull_max": (scipy.stats.weibull_max, {}),
    "gumbel_r": (scipy.stats.gumbel_r, {}),
    "gumbel_l": (scipy.stats.gumbel_l, {}),
    "exponweib": (scipy.stats.exponweib, {}),
    "exponpow": (scipy.stats.exponpow, {}),
    "gamma": (scipy.stats.gamma, {}),
    "logistic": (scipy.stats.logistic, {}),
    "norminvgauss": (scipy.stats.norminvgauss, {}),
    "kstwobign": (scipy.stats.kstwobign, {}),
    "halfnorm": (scipy.stats.halfnorm, {}),
    "halfnormal": (scipy.stats.halfnorm, {}),
    "fatiguelife": (scipy.stats.fatiguelife, {}),
    "nakagami": (scipy.stats.nakagami, {}),
    "wald": (scipy.stats.wald, {}),
    "gompertz": (scipy.stats.gompertz, {}),
    "genextreme": (scipy.stats.genextreme, {}),
    "levy": (scipy.stats.levy, {}),
    "levy_stable": (scipy.stats.levy_stable, {}),
    "laplace": (scipy.stats.laplace, {}),
    "loggamma": (scipy.stats.loggamma, {}),
    "vonmises": (scipy.stats.vonmises, {}),
    "pareto": (scipy.stats.pareto, {}),
    "powerlaw": (scipy.stats.powerlaw, {}),
    "rayleigh": (scipy.stats.rayleigh, {}),
    "rice": (scipy.stats.rice, {}),
    "semicircular": (scipy.stats.semicircular, {}),
    "triang": (scipy.stats.triang, {}),
    "truncexpon": (scipy.stats.truncexpon, {}),
    "truncnorm": (scipy.stats.truncnorm, {}),
    "truncnormal": (scipy.stats.truncnorm, {}),
    "tukeylambda": (scipy.stats.tukeylambda, {}),
    "uniform": (scipy.stats.uniform, {}),
    "wrapcauchy": (scipy.stats.wrapcauchy, {}),
    
    # Discrete Distributions
    "binom": (scipy.stats.binom, {}),
    "bernoulli": (scipy.stats.bernoulli, {}),
    "geom": (scipy.stats.geom, {}),
    "hypergeom": (scipy.stats.hypergeom, {}),
    "poisson": (scipy.stats.poisson, {}),
    "skellam": (scipy.stats.skellam, {}),
    "nbinom": (scipy.stats.nbinom, {}),
    "logser": (scipy.stats.logser, {}),
    "planck": (scipy.stats.planck, {}),
    "boltzmann": (scipy.stats.boltzmann, {}),
    "randint": (scipy.stats.randint, {}),
    "zipf": (scipy.stats.zipf, {}),
    "dlaplace": (scipy.stats.dlaplace, {}),
    "yulesimon": (scipy.stats.yulesimon, {}),
    "hypergeom": (scipy.stats.hypergeom, {}),
    "betabinom": (scipy.stats.betabinom, {}),
    "nbinom": (scipy.stats.nbinom, {}),
    "binom": (scipy.stats.binom, {}),
    "randint": (scipy.stats.randint, {}),
    "boltzmann": (scipy.stats.boltzmann, {}),
    "planck": (scipy.stats.planck, {}),
    "logser": (scipy.stats.logser, {}),
    "zipf": (scipy.stats.zipf, {}),
    "dlaplace": (scipy.stats.dlaplace, {}),
    "skellam": (scipy.stats.skellam, {}),
}



import os
import glob
import click
import xarray as xr

# import own packages
from pymob.utils.store_file import (
    prepare_casestudy, import_package, parse_config_section)
from pymob.utils import help

@click.command()
@click.option("-c", "--case_study", type=str, default="core_daphnia", 
              help=help.case_study)
@click.option("-s", "--scenario", type=str, default="expo_control_2", 
              help=help.scenario)
@click.option("-p", "--posterior_cluster", type=str, default="*")
def main(case_study, scenario, posterior_cluster):

    # parse function parameters
    config = prepare_casestudy((case_study, scenario), "sbi.cfg")

    # import case study and load necessary objects
    mod = import_package(package_path=config["case-study"]["package"])
    plot = getattr(mod.plot, config["sbi"]["posterior_plot"])
    output = config["case-study"]["output"]
    dataset = getattr(mod.data, config["sbi"]["dataset"])
    
    dataset_kwargs = parse_config_section(config["dataset"], "list")
    true_data = dataset(**dataset_kwargs)

    pc = "*" if posterior_cluster == 0 else str(posterior_cluster)

    clusters = glob.glob(os.path.join(output, f"posterior_{pc}.nc"))

    for i in range(1, len(clusters) + 1):
        obsfiles = glob.glob(os.path.join(
            output, "posterior_predictions",f"pred_cluster_{i}_sample_*.nc"
        ))
        obs_list = []        
        for of in obsfiles:
            obs = xr.load_dataarray(of)
            obs_list.append(obs)

        
        # plot the observations
        fig = plot(obs_list, true_data)

        if not isinstance(fig, list):
            fig = [fig]

        for fi, f in enumerate(fig):
            f.savefig(
                os.path.join(config["case-study"]["plots"], 
                f"life_history_cluster_{i}_{fi}.jpg"), dpi=300)

if __name__ == "__main__":
    main()
